# -*- coding: utf-8 -*-
"""
This module contains functionality for working with automated
goal-oriented error control. In particular it offers the following
function:

  compile_with_error_control  - Compile forms and generate error control forms

"""
from .errorcontrol import compile_with_error_control
